{# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Author: Aniek Roelofs -#}
MERGE
  `{{ project_id }}.{{ dataset }}.{{ main_table }}` M
USING
  (SELECT {{ merge_condition_field }} AS id FROM `{{ project_id }}.{{ dataset }}.{{ partitioned_table }}` WHERE _PARTITIONDATE = '{{ ingestion_date }}') P
ON
  M.{{ merge_condition_field }} = P.id
WHEN MATCHED THEN
  DELETE