#!/usr/bin/env bash
# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Tuan Chien


########################################
# Install editable mode Python packages
########################################

{% for package in config.python_packages %}
{% if package.name == 'observatory-api' %}
{% if package.type == 'editable' -%}
cd /opt/{{ package.name }}
export PBR_VERSION=0.0.1
pip3 install -e . --user
unset PBR_VERSION
{% endif %}
{% endif %}
{% endfor %}

# Create database if it does not exist
psql ${BASE_DB_URI} -tc "SELECT 1 FROM pg_database WHERE datname = '${API_SERVER_DB}'" | grep -q 1 || psql ${BASE_DB_URI} -c "CREATE DATABASE ${API_SERVER_DB}"

# Launch api server
gunicorn -b 0.0.0.0:${OBSERVATORY_API_PORT} --timeout 0 observatory.api.server.app:app