{# Copyright 2021 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Author: Aniek Roelofs -#}
# {{ workflow_class }}

Add Workflow Description here

 ```eval_rst
+------------------------------+-----------------------------------------+
| Summary                      |                                         |
+==============================+=========================================+
| Average runtime              | <1h/1-6h/6-12h/12-24h/>24h              |
+------------------------------+-----------------------------------------+
| Average download size        | <100MB/100MB-1GB/1-10GB/10-100GB/>100GB |
+------------------------------+-----------------------------------------+
| Harvest Type                 | API/URL/Other                           |
+------------------------------+-----------------------------------------+
| Workflow Update Frequency    | Daily/Weekly/Monthly/Other              |
+------------------------------+-----------------------------------------+
| Runs on remote worker        | True/False                              |
+------------------------------+-----------------------------------------+
| Catchup missed runs          | True/False                              |
+------------------------------+-----------------------------------------+
| Table Write Disposition      | Truncate/Append                         |
+------------------------------+-----------------------------------------+
| Provider Update Frequency    | Daily/Weekly/Monthly/Other              |
+------------------------------+-----------------------------------------+
| Credentials Required         | Yes/No                                  |
+------------------------------+-----------------------------------------+
| Uses Workflow Template       | Workflow/Snapshot/Stream/Organisation   |
+------------------------------+-----------------------------------------+
| Each shard includes all data | Yes/No                                  |
+------------------------------+-----------------------------------------+
```

## Airflow connections
Note that all values need to be urlencoded.
In the config.yaml file, the following airflow connections are required:

### my_connection
```yaml
my_connection: http://my-username:my-password@
```

## Airflow variables
In the config.yaml file, the following airflow variables are required:

### airflow_variable
```yaml
my_variable_name: my-variable-value
```

## Latest schema
``` eval_rst
.. csv-table::
   :file: ../schemas/{{ workflow_module }}_latest.csv
   :width: 100%
   :header-rows: 1
```