# The backend type: local
# The environment type: develop, staging or production
backend:
  type: local
  environment: develop

# Observatory settings
observatory:
  package: observatory-platform
  package_type: pypi
  airflow_fernet_key: {{ airflow_fernet_key }}
  airflow_secret_key: {{ airflow_secret_key }}

# Terraform settings: customise to use the vm_create and vm_destroy DAGs:
# terraform:
#   organization: my-terraform-org-name

# Google Cloud settings: customise to use Google Cloud services
# google_cloud:
#   project_id: my-gcp-id
#   credentials: /path/to/google_application_credentials.json
#   data_location: us
#   buckets:
#     download_bucket: my-download-bucket-name
#     transform_bucket: my-transform-bucket-name

# User defined Apache Airflow variables:
# airflow_variables:
#   my_variable_name: my-variable-value

# User defined Apache Airflow Connections:
# airflow_connections:
#   my_connection: http://my-username:my-password@

# User defined workflows projects:
# workflows_projects:
#   - package_name: academic-observatory-workflows
#     package: /path/to/academic-observatory-workflows
#     package_type: editable
#     dags_module: academic_observatory_workflows.dags
