{# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose #}

SELECT DISTINCT
  PARSE_DATE('%Y%m%d', _TABLE_SUFFIX) as suffix
FROM
  `{{ project_id }}.{{ dataset_id }}.{{ table_id }}*`
WHERE CHAR_LENGTH(_TABLE_SUFFIX) = 8 AND REGEXP_CONTAINS(_TABLE_SUFFIX, r"\d{8}") AND PARSE_DATE('%Y%m%d', _TABLE_SUFFIX) <= DATE('{{ end_date }}')
ORDER BY suffix DESC
LIMIT {{ limit }};